/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

/**
 * @brief Anything entity-related.
 * @file caosdb/entity.h
 * @author Timm Fitchen
 * @date 2021-07-07
 */
#ifndef CAOSDB_ENTITY_H
#define CAOSDB_ENTITY_H

#include <string>                           // for string
#include "caosdb/entity/v1alpha1/main.pb.h" // for RepeatedPtrField, Message
#include "caosdb/message_code.h"            // for get_message_code, Messag...
#include "google/protobuf/util/json_util.h" // for MessageToJsonString, Jso...

namespace caosdb::entity {
using caosdb::entity::v1alpha1::IdResponse;
using ProtoParent = caosdb::entity::v1alpha1::Parent;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;

/**
 * Parent of an Entity.
 */
class Parent {
public:
  explicit inline Parent(caosdb::entity::v1alpha1::Parent *wrapped)
    : wrapped(wrapped){};
  Parent();
  [[nodiscard]] auto GetId() const -> const std::string &;
  [[nodiscard]] auto GetName() const -> const std::string &;
  [[nodiscard]] auto GetDescription() const -> const std::string &;
  auto SetId(const std::string &id) -> void;
  auto SetName(const std::string &name) -> void;

  inline auto ToString() const -> const std::string {
    google::protobuf::util::JsonOptions options;
    std::string out;
    google::protobuf::util::MessageToJsonString(*(this->wrapped), &out,
                                                options);
    return out;
  }

  friend class Parents;

private:
  static auto CreateProtoParent() -> ProtoParent *;
  mutable caosdb::entity::v1alpha1::Parent *wrapped;
};

/**
 * Container for parents of entities.
 */
class Parents {
public:
  inline Parents(){};
  explicit inline Parents(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Parent>
      *wrapped)
    : wrapped(wrapped){};
  auto Append(const Parent &parent) -> void;
  [[nodiscard]] inline auto Size() const -> int { return wrapped->size(); }
  [[nodiscard]] inline auto At(int index) const -> const Parent {
    return Parent(&(wrapped->at(index)));
  }

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Parent>
    *wrapped;
  friend class Entity;
};

/**
 * Property of an Entity.
 *
 * This is a property which belongs to another entity. Don't confuse it with
 * an Entity with the "Property" role.
 *
 * @brief Property of an Entity.
 */
class Property {
public:
  explicit inline Property(caosdb::entity::v1alpha1::Property *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto GetId() const -> const std::string &;
  [[nodiscard]] auto GetName() const -> const std::string &;
  [[nodiscard]] auto GetDescription() const -> const std::string &;
  [[nodiscard]] auto GetImportance() const -> const std::string &;
  [[nodiscard]] auto GetValue() const -> const std::string &;
  [[nodiscard]] auto GetUnit() const -> const std::string &;
  [[nodiscard]] auto GetDatatype() const -> const std::string &;
  auto SetId(const std::string &id) -> void;
  auto SetName(const std::string &name) -> void;
  auto SetImportance(const std::string &importance) -> void;
  auto SetValue(const std::string &value) -> void;
  auto SetUnit(const std::string &unit) -> void;
  auto SetDatatype(const std::string &datatype) -> void;

private:
  caosdb::entity::v1alpha1::Property *wrapped;
};

/**
 * Container for Properties of Entities
 */
class Properties {
public:
  inline Properties(){};
  explicit inline Properties(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Property>
      *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] auto At(int index) const -> const Property &;
  auto Append(const Property &property) -> void;

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Property>
    *wrapped;
  friend class Entity;
};

class Message {
public:
  explicit inline Message(caosdb::entity::v1alpha1::Message *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] inline auto GetCode() const -> MessageCode {
    return get_message_code(wrapped->code());
  }
  [[nodiscard]] inline auto GetDescription() const -> std::string {
    return wrapped->description();
  }

private:
  caosdb::entity::v1alpha1::Message *wrapped;
};

/**
 * Container for Messages.
 */
class Messages {
public:
  inline Messages() : wrapped(nullptr){};
  explicit inline Messages(
    ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Message>
      *wrapped)
    : wrapped(wrapped){};
  [[nodiscard]] inline auto Size() const -> int { return wrapped->size(); }
  [[nodiscard]] inline auto At(int index) const -> const Message {
    return Message(&(wrapped->at(index)));
  }

private:
  ::google::protobuf::RepeatedPtrField<caosdb::entity::v1alpha1::Message>
    *wrapped;
  friend class Entity;
};

/**
 * @brief Wrapper for the Protobuf entity.
 */
class Entity {
public:
  Entity();
  explicit Entity(IdResponse *idResponse);
  explicit inline Entity(ProtoEntity *wrapped) : wrapped(wrapped) {
    errors.wrapped = this->wrapped->mutable_errors();
    properties.wrapped = this->wrapped->mutable_properties();
    parents.wrapped = this->wrapped->mutable_parents();
  };

  [[nodiscard]] inline auto GetId() const -> const std::string & {
    return wrapped->id();
  };
  [[nodiscard]] inline auto GetVersionId() const -> const std::string & {
    return wrapped->version().id();
  };

  [[nodiscard]] inline auto GetRole() const -> const std::string & {
    return wrapped->role();
  };
  [[nodiscard]] inline auto GetName() const -> const std::string & {
    return wrapped->name();
  };
  [[nodiscard]] inline auto GetDescription() const -> const std::string & {
    return wrapped->description();
  };

  [[nodiscard]] inline auto GetDatatype() const -> const std::string & {
    return wrapped->datatype();
  };
  [[nodiscard]] inline auto GetUnit() const -> const std::string & {
    return wrapped->unit();
  };

  [[nodiscard]] auto GetParents() const -> const Parents &;
  [[nodiscard]] auto GetProperties() const -> const Properties &;
  [[nodiscard]] inline auto GetErrors() const -> const Messages & {
    return errors;
  }
  [[nodiscard]] inline auto HasErrors() const -> bool {
    return this->errors.wrapped->size() > 0;
  }

  inline auto ToString() const -> const std::string {
    google::protobuf::util::JsonOptions options;
    std::string out;
    google::protobuf::util::MessageToJsonString(*(this->wrapped), &out,
                                                options);
    return out;
  }

  auto SetId(const std::string &id) -> void;
  auto SetName(const std::string &name) -> void;
  auto SetVersionId(const std::string &id) -> void;
  auto SetValue(const std::string &value) -> void;
  auto SetUnit(const std::string &unit) -> void;
  auto SetDatatype(const std::string &datatype) -> void;
  auto AppendProperty(const Property &property) -> void;
  auto AppendParent(const Parent &parent) -> void;
  auto Switch(ProtoEntity *entity) -> void;

private:
  static auto CreateProtoEntity() -> ProtoEntity *;
  ProtoEntity *wrapped;
  Properties properties;
  Parents parents;
  Messages errors;
};

} // namespace caosdb::entity
#endif
